#ifndef PLS_TASK_H
#define PLS_TASK_H

#include "pls/internal/scheduling/continuation.h"
#include "pls/internal/scheduling/cont_manager.h"

namespace pls {
namespace internal {
namespace scheduling {

/**
 * A task to be executed by the runtime system.
 * Tasks are guaranteed to be executed exactly once.
 *
 * Override the execute_internal() method for your custom code.
 */
class base_task {
 protected:
  base_task() = default;

  /**
   * Overwrite this with the actual behaviour of concrete tasks.
   */
  virtual void execute_internal() = 0;

 public:
  /**
   * Executes the task and stores its result in the correct continuation.
   * The caller must handle finishing the continuation/informing it that task two was finished.
   */
  void execute() {
    execute_internal();
  }
};

template<typename F>
class task : public base_task {
 public:
  template<typename FARG>
  explicit task(FARG &&function, continuation_node *continuation_node)
      : base_task{}, function_{std::forward<FARG>(function)}, continuation_node_{continuation_node} {}

  void execute_internal() override {
    continuation_node_->get_continuation()->store_result<decltype(function_())>(1, function_());
  }

 private:
  F function_;
  continuation_node *continuation_node_;
};

}
}
}

#endif //PLS_TASK_H
