
#ifndef PLS_TASK_MANAGER_H_
#define PLS_TASK_MANAGER_H_

#include <memory>
#include <utility>
#include <array>
#include <atomic>

#include "pls/internal/scheduling/task.h"
#include "pls/internal/data_structures/bounded_ws_deque.h"
#include "pls/internal/data_structures/stamped_integer.h"

namespace pls {
namespace internal {
namespace scheduling {

struct task_handle {
 public:
  task_handle() : task_{nullptr} {};
  explicit task_handle(base_task *task) : task_{task} {};
  base_task *task_;
};

/**
 * Handles management of tasks in the system. Each thread has a local task manager,
 * responsible for allocating, freeing and publishing tasks for stealing.
 */
class task_manager {
 public:
  // Publishes a task on the stack, i.e. makes it visible for other threads to steal.
  void publish_task(base_task *task) {
    task_deque_.push_bottom(task_handle{task});
  }

  // Try to pop a local task from this task managers stack.
  bool steal_local_task() {
    return task_deque_.pop_bottom();
  }

  // Try to steal a task from a remote task_manager instance. The stolen task must be stored locally.
  // Returns a pair containing the actual task and if the steal was successful.
  // TODO: Re-implement after fast path is done
//  std::pair<task, bool> steal_remote_task(task_manager &other);

  explicit task_manager(data_structures::bounded_ws_deque<task_handle> &task_deque) : task_deque_{task_deque} {}

 private:
  data_structures::bounded_ws_deque<task_handle> &task_deque_;
};

template<size_t NUM_TASKS, size_t MAX_STACK_SIZE>
class static_task_manager {
 public:
  static_task_manager() : task_deque_{}, task_manager_{task_deque_.get_deque()} {};
  task_manager &get_task_manager() { return task_manager_; }

 private:
  data_structures::static_bounded_ws_deque<task_handle, NUM_TASKS> task_deque_;
  task_manager task_manager_;
};

}
}
}
#endif //PLS_TASK_MANAGER_H_
